package com.perforce.hive;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import org.apache.commons.fileupload.FileUploadException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.perforce.hive.api.DeployInterface;
import com.perforce.hive.api.Loader;
import com.perforce.hive.api.artifact.ArtifactInterface;
import com.perforce.hive.api.artifact.ArtifactType;
import com.perforce.hive.api.artifact.ResultInterface;
import com.perforce.hive.api.artifact.SummaryInterface;
import com.perforce.hive.api.repository.RepoInterface;
import com.perforce.hive.api.repository.SiteType;
import com.perforce.hive.api.search.ResultDetailed;
import com.perforce.hive.common.AbstractControl;
import com.perforce.hive.depot.DepotFactory;
import com.perforce.hive.depot.DepotModel;
import com.perforce.hive.repository.RepoFormID;
import com.perforce.hive.site.GroupSite;
import com.perforce.hive.site.ProxySite;
import com.perforce.p4java.core.IDepot;
import com.perforce.p4java.core.file.IFileSpec;
import com.perforce.p4java.exception.P4JavaException;
import com.perforce.p4java.server.IOptionsServer;

import freemarker.cache.ClassTemplateLoader;
import freemarker.cache.MultiTemplateLoader;
import freemarker.cache.TemplateLoader;
import freemarker.template.Configuration;
import ro.fortsoft.pf4j.DefaultPluginManager;
import ro.fortsoft.pf4j.PluginManager;
import spark.Request;
import spark.Response;

public class PluginArtifactFactory extends AbstractControl {

	private static Logger logger = LoggerFactory.getLogger(PluginArtifactFactory.class);

	private static boolean loaded = false;
	private static List<Loader> loaders;
	private static List<ArtifactType> types = new ArrayList<>();
	private static PluginManager pluginManager;
	private static Configuration configuration;

	/**
	 * Load Artifacts
	 * 
	 * @throws IOException
	 */
	public static void loadPlugins() throws IOException {

		if (loaded) {
			logger.warn("Plugins already loaded");
			return;
		}

		pluginManager = new DefaultPluginManager();
		pluginManager.loadPlugins();
		pluginManager.startPlugins();

		loaders = pluginManager.getExtensions(Loader.class);
		for (Loader l : loaders) {
			String plugin = l.getType().getPluginID();
			String id = l.getType().getID();
			logger.info("... loading: " + id + " (" + plugin + ")");
			if (!"UNKNOWN".equals(id)) {
				types.add(l.getType());
			}
			l.loadImage("/images/" + l.getType().getName() + ".png");
		}

		// Load FreeMarker Templates and generate configuration
		loadTemplates();
	}
	
	public static void loadPlugins(File pluginArchiveFile) throws IOException {

		if (loaded) {
			logger.warn("Plugins already loaded");
			return;
		}

		pluginManager = new DefaultPluginManager();
		pluginManager.loadPlugin(pluginArchiveFile);
		pluginManager.startPlugins();

		loaders = pluginManager.getExtensions(Loader.class);
		for (Loader l : loaders) {
			String plugin = l.getType().getPluginID();
			String id = l.getType().getID();
			logger.info("... loading: " + id + " (" + plugin + ")");
			if (!"UNKNOWN".equals(id)) {
				types.add(l.getType());
			}
			l.loadImage("/images/" + l.getType().getName() + ".png");
		}

		// Load FreeMarker Templates and generate configuration
		loadTemplates();
	}

	public static Configuration getConfiguration() {
		return configuration;
	}

	private static void loadTemplates() throws IOException {

		List<TemplateLoader> loaders = new ArrayList<>();
		ClassTemplateLoader template = new ClassTemplateLoader(PluginArtifactFactory.class, "");
		loaders.add(template);

		List<DeployInterface> pages = pluginManager.getExtensions(DeployInterface.class);
		for (DeployInterface p : pages) {
			TemplateLoader loader = p.getTemplateLoader();
			loaders.add(loader);
		}

		// Generate Freemarker configuration
		TemplateLoader[] array = loaders.toArray(new TemplateLoader[loaders.size()]);
		MultiTemplateLoader multi = new MultiTemplateLoader(array);
		configuration = new Configuration();
		configuration.setTemplateLoader(multi);

		for (DeployInterface p : pages) {
			p.addDeployEndPoints("/deploy", configuration);
			p.addAPIEndPoints("/api", configuration);
		}
	}

	public static List<ArtifactType> getTypes() {
		return types;
	}

	public static ArtifactType parseType(String typeStr) {
		for (Loader l : loaders) {
			if (l.getType().getName().equalsIgnoreCase(typeStr)) {
				return l.getType();
			}
		}
		ArtifactType unknown = new ArtifactType(null, "UNSUPPORTED");
		unknown.setTitle("Unknown Artifact");
		return unknown;
	}

	/**
	 * FACTORY: Get Artifact implementation.
	 * 
	 * @param request
	 * @param response
	 * @return
	 * @throws P4JavaException
	 * @throws FileUploadException
	 */
	public static ArtifactInterface getArtifact(ArtifactType type, Request request, Response response)
			throws Exception {
		for (Loader l : loaders) {
			if (l.getType().equals(type)) {
				return l.getArtifact(request, response);
			}
		}
		return null;
	}

	public static ArtifactInterface getProxy(IOptionsServer p4, String repoID) throws Exception {
		DepotModel depot = DepotFactory.getByName(p4, repoID);
		ArtifactType type = depot.getType();

		for (Loader l : loaders) {
			if (l.getType().equals(type)) {
				return l.getProxy(p4, repoID);
			}
		}
		return null;
	}

	public static ResultInterface getResult(ArtifactType type, ResultDetailed result) {
		for (Loader l : loaders) {
			if (l.getType().equals(type)) {
				return l.getResult(result);
			}
		}
		return null;
	}

	public static SummaryInterface getSummary(IOptionsServer p4, ArtifactType type, List<IFileSpec> summary) {
		for (Loader l : loaders) {
			if (l.getType().equals(type)) {
				return l.getSummary(p4, summary);
			}
		}
		return null;
	}

	public static RepoInterface create(Request request) throws Exception {
		String siteStr = RepoFormID.SITE.get(request);
		SiteType site = SiteType.parse(siteStr);

		String typeStr = RepoFormID.TYPE.get(request);
		ArtifactType type = PluginArtifactFactory.parseType(typeStr);

		switch (site) {
		case STORE:
			for (Loader l : loaders) {
				if (l.getType().equals(type)) {
					return l.createRepository(request);
				}
			}
			return null;

		case PROXY:
			return new ProxySite(request);

		case GROUP:
			return new GroupSite(request);

		default:
			return null;
		}

	}

	public static RepoInterface createStore(ArtifactType type) {
		for (Loader l : loaders) {
			if (l.getType().equals(type)) {
				return l.createRepository(type);
			}
		}
		return null;
	}

	public static RepoInterface createRepository(ArtifactType type, IDepot depot) {
		for (Loader l : loaders) {
			if (l.getType().equals(type)) {
				return l.createRepository(depot);
			}
		}
		return null;
	}

	public static Class<?> getRepositoryClass(ArtifactType type) {
		for (Loader l : loaders) {
			if (l.getType().equals(type)) {
				return l.getRepositoryClass();
			}
		}
		return null;
	}
}
