package com.perforce.hive.common;

import com.perforce.hive.connection.ConnectionConfig;
import com.perforce.hive.connection.ConnectionSession;
import com.perforce.p4java.client.IClient;
import com.perforce.p4java.core.IChangelist;
import com.perforce.p4java.core.IChangelistSummary;
import com.perforce.p4java.core.file.FileSpecBuilder;
import com.perforce.p4java.core.file.IFileSpec;
import com.perforce.p4java.exception.P4JavaException;
import com.perforce.p4java.impl.mapbased.client.Client;
import com.perforce.p4java.option.client.RevertFilesOptions;
import com.perforce.p4java.option.server.GetChangelistsOptions;
import com.perforce.p4java.server.IOptionsServer;
import org.apache.commons.io.FileUtils;
import org.apache.commons.pool2.BasePooledObjectFactory;
import org.apache.commons.pool2.PooledObject;
import org.apache.commons.pool2.impl.DefaultPooledObject;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

public class ClientFactory extends BasePooledObjectFactory<ClientPooledObject> {

	private static Logger logger = LoggerFactory.getLogger(ClientFactory.class);
	private static int id = 0;

	private IOptionsServer p4;

	public ClientFactory() {
		String user = ConnectionConfig.get().getProxyUser();
		String ticket = ConnectionConfig.get().getProxyTicket();
		try {
			p4 = ConnectionSession.getP4(user, ticket);
		} catch (P4JavaException e) {
			logger.error("Unable to create pool, invalid Perforce connection.");
			e.printStackTrace();
		}
	}

	@Override
	public ClientPooledObject create() throws Exception {
		// generate unique name
		id++;
		String prefix = ConnectionConfig.get().getClientPrefix();
		String user = p4.getUserName();
		String clientName = prefix + user + "." + id;
		logger.info("client pool (create): " + clientName);

		// create client
		Client implClient = new Client(p4);
		implClient.setName(clientName);
		p4.createClient(implClient);
		IClient iclient = p4.getClient(clientName);
		iclient.setOwnerName(user);

		// set root
		File root = new File("upload", clientName);
		iclient.setRoot(root.getAbsolutePath());
		iclient.update();

		// clean workspace
		cleanWorkspace(iclient);
		return new ClientPooledObject(iclient);
	}

	@Override
	public PooledObject<ClientPooledObject> wrap(ClientPooledObject clientObject) {
		return new DefaultPooledObject<>(clientObject);
	}

	@Override
	public void activateObject(PooledObject<ClientPooledObject> clientObject) throws Exception {
		IClient iclient = clientObject.getObject().getObject();
		cleanWorkspace(iclient);
	}

	@Override
	public void passivateObject(PooledObject<ClientPooledObject> clientObject) throws Exception {
		IClient iclient = clientObject.getObject().getObject();
		logger.info("client pool (passive): " + iclient.getName());

		clean(p4, iclient);
		cleanWorkspace(iclient);
		//p4.deleteClient(client.getName(), false);
	}

	// TODO move to ClientHelper
	public static void clean(IOptionsServer p4, IClient client) throws P4JavaException {
		if (client != null) {
			p4.setCurrentClient(client);

			// revert all opened files
			RevertFilesOptions rOpts = new RevertFilesOptions();
			String path = client.getRoot() + "/...";
			List<IFileSpec> files = FileSpecBuilder.makeFileSpecList(path);
			client.revertFiles(files, rOpts);

			// list pending changes
			List<IFileSpec> empty = new ArrayList<>();
			GetChangelistsOptions opts = new GetChangelistsOptions();
			opts.setType(IChangelist.Type.PENDING);
			opts.setUserName(client.getOwnerName());
			opts.setClientName(client.getName());
			List<IChangelistSummary> pending = p4.getChangelists(empty, opts);

			// deletes pending changes
			for (IChangelistSummary p : pending) {
				p4.deletePendingChangelist(p.getId());
			}
		}
	}

	private void cleanWorkspace(IClient client) throws IOException {
		if(client == null) {
			logger.warn("Client not defined.");
			return;
		}

		File root = new File(client.getRoot());
		if (root.exists()) {
			FileUtils.deleteDirectory(root);
		}
		if (!root.exists() && !root.mkdirs()) {
			logger.warn("Failed to create upload directory.");
		}
	}
}
