package com.perforce.hive.depot;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.perforce.p4java.admin.IProtectionEntry;
import com.perforce.p4java.exception.P4JavaException;
import com.perforce.p4java.impl.generic.admin.ProtectionEntry;
import com.perforce.p4java.server.CmdSpec;
import com.perforce.p4java.server.IOptionsServer;

public class DepotProtect {

	private static Logger logger = LoggerFactory.getLogger(DepotProtect.class);

	private final IOptionsServer p4;

	private Map<String, List<IProtectionEntry>> map = new HashMap<>();
	private List<IProtectionEntry> top = new ArrayList<>();
	private List<IProtectionEntry> bottom = new ArrayList<>();

	public DepotProtect(IOptionsServer p4) throws P4JavaException {
		this.p4 = p4;
		getEntries();
	}

	public void setPermissions(String depot, IProtectionEntry[] array) {
		// exit early if no permissions are set
		if (array == null) {
			return;
		}
		List<IProtectionEntry> list = new ArrayList<>();
		for (IProtectionEntry a : array) {
			list.add(a);
		}
		map.put(depot, list);
	}

	public List<IProtectionEntry> getPermissions(String depot) {
		if (map.containsKey(depot)) {
			return map.get(depot);
		} else {
			return new ArrayList<>();
		}
	}

	public void save() throws P4JavaException {
		List<IProtectionEntry> list = new ArrayList<>();
		int index = 0;

		// Add original protections above HIVE entry
		for (IProtectionEntry e : top) {
			e.setOrder(index);
			list.add(e);
			index++;
		}

		// Add HIVE protections
		for (Entry<String, List<IProtectionEntry>> depot : map.entrySet()) {
			for (IProtectionEntry e : depot.getValue()) {
				e.setOrder(index);
				list.add(e);
				index++;
			}
		}

		// Add original protections below HIVE entry
		for (IProtectionEntry e : bottom) {
			e.setOrder(index);
			list.add(e);
			index++;
		}

		String status = p4.updateProtectionEntries(list);
		logger.info(status);
	}

	@Override
	public String toString() {
		StringBuffer sb = new StringBuffer();
		for (Map.Entry<String, List<IProtectionEntry>> entry : map.entrySet()) {
			String depot = entry.getKey();
			sb.append("Depot: " + depot + "\n");

			List<IProtectionEntry> list = entry.getValue();
			for (IProtectionEntry e : list) {
				sb.append("... ");
				sb.append(e.getOrder() + " ");
				sb.append(e.getMode() + " ");
				sb.append(e.isGroup() ? "group" : "user" + " ");
				sb.append(e.getType() + " ");
				sb.append(e.getName() + " ");
				sb.append(e.getPath() + " ");
				sb.append("\n");
			}
		}
		return sb.toString();
	}

	private void getEntries() throws P4JavaException {

		List<IProtectionEntry> current = top;
		List<IProtectionEntry> entries = getProtectionsTable();
		List<IProtectionEntry> hive = new ArrayList<>();

		for (IProtectionEntry e : entries) {
			if ("_____HIVE:start_____".equals(e.getPath())) {
				current.add(e);
				current = hive;
			} else if ("_____HIVE:end_____".equals(e.getPath())) {
				current = bottom;
				current.add(e);
			} else {
				current.add(e);
			}
		}

		if (hive.isEmpty()) {
			logger.warn("HIVE protection block not found.");
			return;
		}

		Pattern pattern = Pattern.compile("-?//(.*)/\\.\\.\\.");

		for (IProtectionEntry h : hive) {
			String path = h.getPath();
			Matcher m = pattern.matcher(path);
			m.find();

			String depot = m.group(1);
			List<IProtectionEntry> list;
			if (map.containsKey(depot)) {
				list = map.get(depot);
			} else {
				list = new ArrayList<>();
			}
			list.add(h);

			// sort entries before adding to map
			sort(list);
			map.put(depot, list);
		}
	}

	private List<IProtectionEntry> getProtectionsTable() throws P4JavaException {

		List<IProtectionEntry> list = new ArrayList<>();
		Pattern pattern = Pattern.compile("^(.*) (.*) (.*) (.*) (.*)$");

		List<Map<String, Object>> resultMaps = p4.execMapCmdList(CmdSpec.PROTECT.name(), new String[] { "-o" }, null);

		if (resultMaps == null) {
			return list;
		}

		for (Map<String, Object> map : resultMaps) {
			if (map == null) {
				continue;
			}
			for (Map.Entry<String, Object> entry : map.entrySet()) {
				String key = entry.getKey();
				if (!key.startsWith("Protections")) {
					continue;
				}

				String idxString = key.substring("Protections".length());
				int index = Integer.parseInt(idxString);
				String value = (String) entry.getValue();
				Matcher m = pattern.matcher(value);
				m.find();

				String mode = m.group(1);
				String group = m.group(2);
				String name = m.group(3);
				String host = m.group(4);
				String path = m.group(5);

				boolean isGroup = "group".equals(group);
				boolean isExclude = path.startsWith("-");

				IProtectionEntry protection = new ProtectionEntry(index, mode, isGroup, host, name, path, isExclude);
				list.add(protection);
			}
		}

		// Sort protection entries
		sort(list);

		return list;
	}

	private void sort(List<IProtectionEntry> list) {
		Collections.sort(list, new Comparator<IProtectionEntry>() {
			@Override
			public int compare(IProtectionEntry o1, IProtectionEntry o2) {
				int i1 = o1.getOrder();
				int i2 = o2.getOrder();
				return (i1 > i2) ? 1 : -1;
			}
		});
	}

}
